#include "DirWatch.hpp"

#define TRACE(str) do { puts(str); fflush(stdout); } while (0)

LPCSTR W2A(LPCWSTR psz)
{
    static char s_sz[2][MAX_PATH] = { "" };
    static BOOL s_bFlag = FALSE;
    if (psz == NULL)
        return "(null)";
    WideCharToMultiByte(CP_ACP, 0, psz, -1, s_sz[!s_bFlag], MAX_PATH, NULL, NULL);
    s_bFlag = !s_bFlag;
    return s_sz[s_bFlag];
}

inline std::wstring DoGetFullPath(LPCWSTR pszPath)
{
    WCHAR szPath[MAX_PATH];
    GetFullPathNameW(pszPath, _countof(szPath), szPath, NULL);
    return szPath;
}

std::wstring DoAddBackslash(const std::wstring& strPath)
{
    if (strPath.empty())
        return strPath;
    if (strPath[strPath.size() - 1] == L'\\')
        return strPath;
    std::wstring ret = strPath;
    ret += L'\\';
    return ret;
}

std::wstring DoGetPathName(CDirWatchItem *pWatchItem, FILE_NOTIFY_INFORMATION *pFNI)
{
    WCHAR szPath[MAX_PATH] = L"";
    SIZE_T cbMax = pFNI->FileNameLength;
    if (cbMax > sizeof(szPath))
        cbMax = sizeof(szPath);
    CopyMemory(szPath, pFNI->FileName, cbMax);
    szPath[_countof(szPath) - 1] = 0;

    std::wstring strPath = DoAddBackslash(pWatchItem->m_strDir);
    strPath += szPath;

    return strPath;
}

//////////////////////////////////////////////////////////////////////////////
// CObj

#ifndef NDEBUG
    LONG CObj::s_nAliveCount = 0;
#endif

inline CObj::CObj() : m_cRefs(1)
{
    TRACE("CObj::CObj");
    #ifndef NDEBUG
        ++s_nAliveCount;
    #endif
}

inline CObj::~CObj()
{
    TRACE("CObj::~CObj");
    #ifndef NDEBUG
        --s_nAliveCount;
    #endif
}

inline ULONG CObj::AddRef()
{
    TRACE("CObj::AddRef");
    return ::InterlockedIncrement(&m_cRefs);
}

inline ULONG CObj::Release()
{
    TRACE("CObj::Release");
    ULONG cRefs = ::InterlockedDecrement(&m_cRefs);
    if (cRefs == 0)
    {
        delete this;
        return 0;
    }
    return cRefs;
}

inline LONG CObj::GetRefCount() const
{
    return m_cRefs;
}

//////////////////////////////////////////////////////////////////////////////
// CDirChangeHandler

inline CDirChangeHandler::CDirChangeHandler()
{
}

inline CDirChangeHandler::~CDirChangeHandler()
{
}

BOOL CDirChangeHandler::OnChangeNotify(DWORD dwAction, LPCWSTR file1, LPCWSTR file2)
{
    switch (dwAction)
    {
    case FILE_ACTION_ADDED:
        printf("FILE_ACTION_ADDED: %s, %s\n", W2A(file1), W2A(file2));
        break;
    case FILE_ACTION_REMOVED:
        printf("FILE_ACTION_REMOVED: %s, %s\n", W2A(file1), W2A(file2));
        break;
    case FILE_ACTION_MODIFIED:
        printf("FILE_ACTION_MODIFIED: %s, %s\n", W2A(file1), W2A(file2));
        break;
    case FILE_ACTION_RENAMED_NEW_NAME:
        printf("FILE_ACTION_RENAMED_NEW_NAME: %s, %s\n", W2A(file1), W2A(file2));
        break;
    default:
        printf("%08X: %s, %s\n", dwAction, W2A(file1), W2A(file2));
        break;
    }
    fflush(stdout);
    return TRUE;
}

//////////////////////////////////////////////////////////////////////////////
// CDirWatchItem

CDirWatchItem::CDirWatchItem(LPCWSTR pszDir, DWORD dwChangeFilter, BOOL bWatchSubDir)
    : m_strDir(pszDir)
    , m_hDir(INVALID_HANDLE_VALUE)
    , m_dwChangeFilter(dwChangeFilter)
    , m_bWatchSubDir(bWatchSubDir)
    , m_dwReadDirError(ERROR_SUCCESS)
    , m_hStartEvent(::CreateEvent(NULL, TRUE, FALSE, NULL))
{
    TRACE("CDirWatchItem");
    ZeroMemory(&m_Overlapped, sizeof(m_Overlapped));
    assert(m_hStartEvent);
}

inline CDirWatchItem::~CDirWatchItem()
{
    TRACE("~CDirWatchItem");
    CloseDirHandle();
    CloseStartEvent();
}

BOOL CDirWatchItem::ReopenDir(HANDLE& hCompPort)
{
    CloseDirHandle();
    HANDLE hDir = ::CreateFileW(m_strDir.c_str(),
                                FILE_LIST_DIRECTORY,
                                FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
                                NULL,
                                OPEN_EXISTING,
                                FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OVERLAPPED,
                                NULL);
    if (hDir == INVALID_HANDLE_VALUE)
    {
        assert(0);
        return FALSE;
    }

    hCompPort = ::CreateIoCompletionPort(hDir, hCompPort,
                                         reinterpret_cast<ULONG_PTR>(this), 0);
    if (!hCompPort)
    {
        assert(0);
        CloseHandle(hDir);
        return FALSE;
    }

    m_hDir = hDir;
    ::PostQueuedCompletionStatus(hCompPort,
                                 sizeof(this),
                                 reinterpret_cast<ULONG_PTR>(this),
                                 &m_Overlapped);
    return TRUE;
}

BOOL CDirWatchItem::IsDir()
{
    return this && ::PathIsDirectoryW(m_strDir.c_str());
}

void CDirWatchItem::CloseDirHandle()
{
    TRACE("CloseDirHandle");
    if (m_hDir != INVALID_HANDLE_VALUE)
    {
        CloseHandle(m_hDir);
        m_hDir = INVALID_HANDLE_VALUE;
    }
}

void CDirWatchItem::SetStartEvent()
{
    assert(m_hStartEvent);
    ::SetEvent(m_hStartEvent);
}

BOOL CDirWatchItem::WaitForStartEvent()
{
    DWORD dwWait = 0;
    do
    {
        dwWait = ::WaitForSingleObject(m_hStartEvent, 300);
    } while (dwWait != WAIT_OBJECT_0);
    ::ResetEvent(m_hStartEvent);
    return dwWait == WAIT_OBJECT_0;
}

void CDirWatchItem::CloseStartEvent()
{
    if (m_hStartEvent)
    {
        ::CloseHandle(m_hStartEvent);
        m_hStartEvent = NULL;
    }
}

//////////////////////////////////////////////////////////////////////////////
// CDirChangeWatcher

CDirChangeWatcher::CDirChangeWatcher()
    : m_hCompPort(NULL)
    , m_hThread(NULL)
    , m_dwThreadID(0)
    , m_pChangeHandler(NULL)
    , m_nWatchState(WATCH_STATE_INITIAL)
{
}

CDirChangeWatcher::~CDirChangeWatcher()
{
    UnwatchAll();
    CloseThread();
    CloseCompPort();
    ReleaseChangeHandler();
}

void CDirChangeWatcher::ReleaseChangeHandler()
{
    if (m_pChangeHandler)
    {
        m_pChangeHandler->Release();
        m_pChangeHandler = NULL;
    }
}

void CDirChangeWatcher::CloseCompPort()
{
    if (m_hCompPort)
    {
        ::CloseHandle(m_hCompPort);
        m_hCompPort = NULL;
    }
}

DWORD CDirChangeWatcher::WatchProc()
{
    TRACE("WatchProc");
    BYTE abBuffer[4092] = "";

    assert(m_hCompPort);
    assert(GetCurrentThreadId() == m_dwThreadID);

    DWORD cbStatus = 0;
    CDirWatchItem *pWatchItem = NULL;
    LPOVERLAPPED pOverlapped;

    do
    {
        printf("m_dwThreadID: %u\n", m_dwThreadID);

        if (!::GetQueuedCompletionStatus(m_hCompPort, &cbStatus,
                                         reinterpret_cast<PULONG_PTR>(&pWatchItem),
                                         &pOverlapped, INFINITE))
        {
            TRACE("GetQueuedCompletionStatus failed");
            break;
        }

        Lock();
        WATCH_STATE state = m_nWatchState;
        Unlock();

        switch (state)
        {
        case WATCH_STATE_ADD_ITEM:
            TRACE("WATCH_STATE_ADD_ITEM");

            if (!pWatchItem)
                break;

            if (::ReadDirectoryChangesW(pWatchItem->m_hDir,
                                        abBuffer,
                                        sizeof(abBuffer),
                                        pWatchItem->m_bWatchSubDir,
                                        pWatchItem->m_dwChangeFilter,
                                        NULL,
                                        &pWatchItem->m_Overlapped,
                                        NULL))
            {
                m_nWatchState = WATCH_STATE_WAIT;
                pWatchItem->m_dwReadDirError = ERROR_SUCCESS;
            }
            else
            {
                pWatchItem->m_dwReadDirError = ::GetLastError();
            }
            pWatchItem->SetStartEvent();
            break;

        case WATCH_STATE_STOP:
            TRACE("WATCH_STATE_STOP");
            break;

        case WATCH_STATE_WAIT:
            TRACE("WATCH_STATE_WAIT");

            if (!cbStatus)
                break;

            if (!pWatchItem)
                break;

            Lock();
            ProcessNotifications(pWatchItem, abBuffer, sizeof(abBuffer));
            Unlock();

            if (::ReadDirectoryChangesW(pWatchItem->m_hDir,
                                        abBuffer,
                                        sizeof(abBuffer),
                                        pWatchItem->m_bWatchSubDir,
                                        pWatchItem->m_dwChangeFilter,
                                        NULL,
                                        &pWatchItem->m_Overlapped,
                                        NULL))
            {
                pWatchItem->m_dwReadDirError = ERROR_SUCCESS;
            }
            else
            {
                pWatchItem->m_dwReadDirError = GetLastError();

                INT nOldPriority = ::GetThreadPriority(::GetCurrentThread());
                ::SetThreadPriority(::GetCurrentThread(), THREAD_PRIORITY_HIGHEST);

                UnwatchDir(pWatchItem);

                ::SetThreadPriority(::GetCurrentThread(), nOldPriority);
            }
            break;
        }
    } while (pWatchItem);

    TRACE("WatchProc: quit");
    return 0;
}

void CDirChangeWatcher::ProcessNotifications(CDirWatchItem *pWatchItem, LPBYTE pbBuffer, size_t cbBuffer)
{
    TRACE("ProcessNotifications");

    if (!pWatchItem || !m_pChangeHandler)
    {
        assert(0);
        return;
    }

    LPBYTE pb = pbBuffer;
    size_t ib = 0;
    std::wstring strPath, strOldPath;

    for (;;)
    {
        FILE_NOTIFY_INFORMATION *pFNI = reinterpret_cast<FILE_NOTIFY_INFORMATION*>(pb);

        switch (pFNI->Action)
        {
        case FILE_ACTION_ADDED:
        case FILE_ACTION_REMOVED:
        case FILE_ACTION_MODIFIED:
        case FILE_ACTION_RENAMED_NEW_NAME:
            strPath = DoGetPathName(pWatchItem, pFNI);
            if (pFNI->Action == FILE_ACTION_RENAMED_NEW_NAME)
                m_pChangeHandler->OnChangeNotify(pFNI->Action, strOldPath.c_str(), strPath.c_str());
            else
                m_pChangeHandler->OnChangeNotify(pFNI->Action, strPath.c_str(), NULL);
            break;

        case FILE_ACTION_RENAMED_OLD_NAME:
            strOldPath = DoGetPathName(pWatchItem, pFNI);
            break;
        }

        if (pFNI->NextEntryOffset == 0)
            break;

        pb += pFNI->NextEntryOffset;
        ib += pFNI->NextEntryOffset;
        assert(ib <= cbBuffer);
    }
}

inline void CDirChangeWatcher::CloseThread()
{
    if (m_hThread)
    {
        ::CloseHandle(m_hThread);
        m_hThread = NULL;
        m_dwThreadID = 0;
    }
}

BOOL CDirChangeWatcher::IsWatching()
{
    BOOL ret = FALSE;
    Lock();
    for (size_t iItem = 0; iItem < m_WatchItems.size(); ++iItem)
    {
        if (m_WatchItems[iItem]->IsDir())
        {
            ret = TRUE;
            break;
        }
    }
    Unlock();
    return ret;
}

inline BOOL CDirChangeWatcher::IsWatchingDir(LPCWSTR pszDir)
{
    Lock();
    INT_PTR iItem = FindWatchItem(pszDir);
    Unlock();
    return iItem >= 0;
}

DWORD CDirChangeWatcher::WatchDir(LPCWSTR pszDir, DWORD dwChangeFilter, BOOL bWatchSubDir)
{
    TRACE("WatchDir");

    if (!pszDir || !pszDir[0] || !m_pChangeHandler)
    {
        assert(0);
        return ERROR_INVALID_PARAMETER;
    }

    if (!::PathIsDirectoryW(pszDir))
    {
        assert(0);
        return ERROR_BAD_PATHNAME;
    }

    Lock();
    INT_PTR iItem = FindWatchItem(pszDir);
    Unlock();
    if (iItem >= 0)
    {
        assert(m_WatchItems[iItem]);
        m_WatchItems[iItem]->AddRef();
        return ERROR_SUCCESS;
    }

    CDirWatchItem *pWatchItem = new CDirWatchItem(pszDir, dwChangeFilter, bWatchSubDir);
    if (!IsThreadRunning())
    {
        m_hThread = reinterpret_cast<HANDLE>(
            ::_beginthreadex(NULL, 0, WatchThreadProc, this, 0, &m_dwThreadID));
        if (!m_hThread)
        {
            assert(0);
            return ERROR_MAX_THRDS_REACHED;
        }
    }

    if (AddWatchItem(pWatchItem) < 0)
        return ERROR_INVALID_PARAMETER;
    return ERROR_SUCCESS;
}

INT_PTR CDirChangeWatcher::AddWatchItem(CDirWatchItem *pWatchItem)
{
    assert(pWatchItem);

    m_nWatchState = WATCH_STATE_ADD_ITEM;
    if (!pWatchItem->ReopenDir(m_hCompPort))
    {
        assert(0);
        pWatchItem->Release();
        return -1;
    }
    pWatchItem->WaitForStartEvent();

    if (pWatchItem->m_dwReadDirError != ERROR_SUCCESS)
    {
        assert(0);
        pWatchItem->Release();
        return -1;
    }

    Lock();
    for (size_t iItem = 0; iItem < m_WatchItems.size(); ++iItem)
    {
        if (m_WatchItems[iItem] == NULL)
        {
            m_WatchItems[iItem] = pWatchItem;
            Unlock();
            return iItem;
        }
    }
    size_t iItem = m_WatchItems.size();
    m_WatchItems.push_back(pWatchItem);
    Unlock();
    return iItem;
}

INT_PTR CDirChangeWatcher::FindWatchItemExact(LPCWSTR pszDir) const
{
    for (size_t iItem = 0; iItem < m_WatchItems.size(); ++iItem)
    {
        CDirWatchItem *pWatchItem = m_WatchItems[iItem];
        if (pWatchItem && lstrcmpiW(pWatchItem->m_strDir.c_str(), pszDir) == 0)
            return iItem;
    }
    return -1;
}

INT_PTR CDirChangeWatcher::FindWatchItem(LPCWSTR pszDir) const
{
    for (size_t iItem = 0; iItem < m_WatchItems.size(); ++iItem)
    {
        CDirWatchItem *pWatchItem = m_WatchItems[iItem];
        if (!pWatchItem)
            continue;

        std::wstring strPath1 = DoAddBackslash(pWatchItem->m_strDir);
        std::wstring strPath2 = DoAddBackslash(pszDir);
        if (pWatchItem->m_bWatchSubDir)
        {
            if (strPath1.size() == strPath2.size())
            {
                if (lstrcmpiW(strPath1.c_str(), strPath2.c_str()) == 0)
                    return iItem;
            }
            else if (strPath1.size() < strPath2.size())
            {
                strPath2.resize(strPath1.size());
                if (lstrcmpiW(strPath1.c_str(), strPath2.c_str()) == 0)
                    return iItem;
            }
        }
        else
        {
            if (lstrcmpiW(strPath1.c_str(), strPath2.c_str()) == 0)
                return iItem;
        }
    }
    return -1;
}

INT_PTR CDirChangeWatcher::FindWatchItem(CDirWatchItem *pWatchItem) const
{
    for (size_t iItem = 0; iItem < m_WatchItems.size(); ++iItem)
    {
        if (pWatchItem && m_WatchItems[iItem] == pWatchItem)
            return iItem;
    }
    return -1;
}

BOOL CDirChangeWatcher::UnwatchDir(INT_PTR iItem)
{
    if (iItem < 0 || m_WatchItems.size() <= size_t(iItem))
    {
        assert(0);
        return FALSE;
    }

    CDirWatchItem *pWatchItem = m_WatchItems[iItem];
    if (pWatchItem->GetRefCount() == 1)
    {
        m_WatchItems[iItem] = NULL;
    }
    pWatchItem->Release();

    if (!IsWatching())
    {
        Lock();
        m_nWatchState = WATCH_STATE_STOP;
        ::PostQueuedCompletionStatus(m_hCompPort, 0, 0, NULL);
        CloseThread();
        Unlock();
    }

    return TRUE;
}

BOOL CDirChangeWatcher::UnwatchDir(LPCWSTR pszDir)
{
    if (!m_hCompPort || !pszDir || !pszDir[0])
        return FALSE;

    BOOL ret = FALSE;

    Lock();
    INT_PTR iItem = FindWatchItemExact(pszDir);
    if (iItem < 0)
        iItem = FindWatchItem(pszDir);
    if (iItem >= 0)
        ret = UnwatchDir(iItem);
    Unlock();

    return ret;
}

BOOL CDirChangeWatcher::UnwatchDir(CDirWatchItem *pWatchItem)
{
    BOOL ret = FALSE;
    assert(pWatchItem);

    Lock();
    INT_PTR iItem = FindWatchItem(pWatchItem);
    if (iItem >= 0)
        ret = UnwatchDir(iItem);
    Unlock();

    return ret;
}

BOOL CDirChangeWatcher::IsThreadRunning()
{
    if (!m_hThread)
        return FALSE;

    DWORD dwExitCode;
    if (::GetExitCodeThread(m_hThread, &dwExitCode) && dwExitCode == STILL_ACTIVE)
        return TRUE;

    CloseHandle(m_hThread);
    m_hThread = NULL;
    m_dwThreadID = 0;
    return FALSE;
}

BOOL CDirChangeWatcher::UnwatchAll()
{
    TRACE("UnwatchAll");

    if (!IsThreadRunning())
    {
        CloseCompPort();
        return TRUE;
    }

    Lock();
    assert(m_hCompPort);

    for (size_t iItem = 0; iItem < m_WatchItems.size(); ++iItem)
    {
        CDirWatchItem *pWatchItem = m_WatchItems[iItem];
        if (pWatchItem)
        {
            UnwatchDir(pWatchItem);
            pWatchItem = m_WatchItems[iItem];
            if (pWatchItem)
                delete pWatchItem;
        }
    }
    m_WatchItems.clear();

    ::PostQueuedCompletionStatus(m_hCompPort, 0, 0, NULL);
    CloseCompPort();
    CloseThread();

    Unlock();
    return TRUE;
}

void CDirChangeWatcher::SetChangeHandler(CDirChangeHandler *pChangeHandler)
{
    Lock();
    if (m_pChangeHandler != pChangeHandler)
    {
        if (m_pChangeHandler)
            m_pChangeHandler->Release();
        m_pChangeHandler = pChangeHandler;
        if (m_pChangeHandler)
            m_pChangeHandler->AddRef();
    }
    Unlock();
}

void CDirChangeWatcher::ReopenDirs()
{
    for (size_t iItem = 0; iItem < m_WatchItems.size(); ++iItem)
    {
        CDirWatchItem *pItem = m_WatchItems[iItem];
        if (!pItem)
            continue;
        m_nWatchState = WATCH_STATE_ADD_ITEM;
        pItem->ReopenDir(m_hCompPort);
    }
}

//////////////////////////////////////////////////////////////////////////////
// main

int main(void)
{
    CDirChangeWatcher *pWatcher = new CDirChangeWatcher();
    CDirChangeHandler *pHandler = new CDirChangeHandler();
    BOOL bWatchSubDir = TRUE;

    LPCWSTR pszDir1 = L"mydir1";
    LPCWSTR pszDir2 = L"mydir2";
    DWORD dwChangeFilter =
        FILE_NOTIFY_CHANGE_FILE_NAME |
        FILE_NOTIFY_CHANGE_DIR_NAME |
        FILE_NOTIFY_CHANGE_ATTRIBUTES |
        FILE_NOTIFY_CHANGE_SIZE |
        FILE_NOTIFY_CHANGE_LAST_WRITE |
        FILE_NOTIFY_CHANGE_LAST_ACCESS |
        FILE_NOTIFY_CHANGE_CREATION |
        FILE_NOTIFY_CHANGE_SECURITY;

    pWatcher->SetChangeHandler(pHandler);
    pWatcher->WatchDir(pszDir1, dwChangeFilter, bWatchSubDir);
    //pWatcher->WatchDir(pszDir2, dwChangeFilter, bWatchSubDir);

    for (INT i = 0; i < 10; ++i)
    {
        TRACE("IsWatching 1");
        Sleep(500);
    }

    pWatcher->UnwatchDir(pszDir1);
    //pWatcher->UnwatchAll();
    pWatcher->WatchDir(pszDir2, dwChangeFilter, bWatchSubDir);
    //pWatcher->ReopenDirs();

    for (INT i = 0; i < 10; ++i)
    {
        if (!pWatcher->IsWatching())
            break;
        TRACE("IsWatching 2");
        Sleep(500);
    }

    pWatcher->UnwatchAll();

    pWatcher->Release();
    pHandler->Release();

    TRACE("main: quit");

#if defined(_MSC_VER) && !defined(NDEBUG)
    // for detecting memory leak (MSVC only)
    _CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);
#endif

    assert(CObj::s_nAliveCount == 0);

    return 0;
}
